import sys
from tahoe.tests import TahoeTestCase
from app import TestTahoeApplication

sys.stderr.write("""

     myssshhssssd.    `h.     `/shs/`  `/ohs/.    :ososyy+-   `/ohy++syo
    -s    mM`   s.    sMd       sM/      :My    :do`    `+Nh.   :Mh   .y
          mM`        /m/Ms      -M/      :My   :Ms        :NN.  :Mh   .
          mM`       .N. oM/     -Ms//////sMy   dM.         hM-  :MmssyN
          mM`      `dssssdN.    -M/      :My   yMy         dN.  :Mh   +
          mM`      ys    .NN`   +M/      :My   `hMs`      om:   :Mh    //
        :/NM+/  `/sN+.   `sMd/-/dMd/`  `/sMm/-   -ydhooooy/   `/sMNoooyN.

     [ Tests will free you! FFFFFFFFUUUUU.... ]

""")
sys.stderr.flush()

test_app = TestTahoeApplication()
test_app.setup()
TahoeTestCase.register_tahoe_application(test_app)



import datetime
import decimal

from tahoe.encoding import json_handler
from tahoe.tests import TahoeTestCase


class Dummy(object):
    def __init__(self, first_name, last_name):
        self.first_name = first_name
        self.last_name = last_name

    def to_dict(self):
        return dict(
            first_name=self.first_name,
            last_name=self.last_name)


class TestEncoding(TahoeTestCase):
    def test_json_encoder_uses_to_dict(self):
        """
        JSON Encoder will use the to_dict method of a class instance
        """
        first_name = 'clint`'
        last_name = 'ecker'
        dummy = Dummy(first_name=first_name, last_name=last_name)
        dummy_handled = json_handler(dummy)

        assert isinstance(dummy_handled, dict)
        assert dummy_handled.get('first_name') == first_name
        assert dummy_handled.get('last_name') == last_name

    def test_json_encoder_handles_datetimes(self):
        """
        JSON Encoder handles datetimes
        """
        dt = datetime.datetime.utcnow()
        dt_handled = json_handler(dt)

        assert isinstance(dt_handled, basestring)
        assert '.' not in dt_handled

    def test_json_encoder_handles_dates(self):
        """
        JSON Encoder handles dates
        """
        d = datetime.date.today()
        d_handled = json_handler(d)

        assert isinstance(d_handled, basestring)
        assert ':' not in d_handled

    def test_json_encoder_handles_decimals(self):
        """
        JSON Encoder handles decimals
        """
        f = 3.14159265358979
        d = decimal.Decimal(f)
        d_handled = json_handler(d)
        assert isinstance(d_handled, float)
        assert d_handled == f

    def test_json_encoder_returns_unhandled_objects(self):
        """
        JSON Encoder doesn't mess with unhandled objects
        """
        s = "Clint is Cool"
        s_handled = json_handler(s)
        assert s is s_handled



import base64
import pickle
from collections import defaultdict
from mock import Mock

from tahoe.models import ModelRegistry

from tahoe.models.mixins.encodable import EncoderMixin
from tahoe.models.mixins.searchable import SearchableMixin
from tahoe.models.mixins.timestampable import TimestampMixin

from tahoe import search

from tahoe.tests import TahoeTestCase


class TestEncoderMixin(TahoeTestCase):
    def test_encoder_mixin(self):
        """
        EncoderMixin: Test encoder and decoder chains
        """
        class TestBase(EncoderMixin):
            encode_decode_chain = [
                (pickle.dumps, pickle.loads),
                (base64.b64encode, base64.b64decode)]

        tb = TestBase()

        data = dict(
            one="one",
            two="two",
            three="three")

        encoded_data = tb.encode(data)
        assert encoded_data
        decoded_data = tb.decode(encoded_data)
        assert decoded_data

        assert decoded_data == data
        assert encoded_data != data

    def test_encoder_mixin_fails_with_bad_chain(self):
        """
        EncoderMixin: Decoding and Encoding fail with bad settings
        """
        class TestBadChain(EncoderMixin):
            encode_decode_chain = [
                (base64.b64encode, pickle.loads),
                (pickle.dumps, base64.b64decode)]

        tb = TestBadChain()

        data = dict(
            one="one",
            two="two",
            three="three")

        with self.assertRaises(Exception):
            tb.encode(data)


class TestTimestampMixin(TahoeTestCase):
    @staticmethod
    def make_callback():
        def callback(*callback_args, **callback_kwargs):
            db = callback_kwargs.get('db')

            class TestModel(db.Model,
                            TimestampMixin(callback_kwargs.get('db'))):
                pass
            return TestModel
        return callback

    def setUp(self):
        self.mock_config = Mock(name='config')
        self.mock_app = Mock(name='app')
        self.mock_logger = Mock(name='logger')
        self.mock_bucket = defaultdict(Mock)
        self.mock_query = Mock(name='query')
        self.mock_filter = Mock(name='filter')

        self.mock_query.filter = self.mock_filter

        class ModelClass(object):
            query = self.mock_query

        self.mock_db = Mock(name='db')
        self.mock_db.Model = ModelClass

        ModelRegistry.destroy_all()

    def test_timestamp_mixin(self):
        ModelRegistry.register()(self.make_callback())
        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()
        assert 'TestModel' in mr.models
        assert mr.TestModel

        assert self.mock_db.Column.call_count == 3

        # Verify inhereted class attributes
        assert hasattr(mr.TestModel, 'NOW_FN')
        assert hasattr(mr.TestModel, 'created_at')
        assert hasattr(mr.TestModel, 'updated_at')
        assert hasattr(mr.TestModel, 'deleted_at')
        assert hasattr(mr.TestModel, 'query_active')
        assert hasattr(mr.TestModel, 'query')

        q = mr.TestModel.query_active()

        assert q
        assert self.mock_filter.called
        assert self.mock_filter.called_with(False)

        # Instance attributes
        tm = mr.TestModel()
        assert hasattr(tm, 'is_deleted')
        assert hasattr(tm, 'mark_deleted')
        assert hasattr(tm, 'mark_undeleted')

        tm.mark_undeleted()
        assert not tm.is_deleted()
        tm.mark_deleted()
        assert tm.is_deleted()


class TestSearchableMixin(TahoeTestCase):
    @staticmethod
    def make_callback(bad=False):
        def callback(*callback_args, **callback_kwargs):
            bucket = callback_kwargs.get('bucket')
            logger = callback_kwargs.get('logger')
            db = callback_kwargs.get('db')

            class TestModel(db.Model,
                            SearchableMixin(bucket=bucket, logger=logger)):
                @classmethod
                def _search_index_mapping(cls):
                    return dict(properties=dict(
                        id=search.EXACT_MATCH_FIELD,
                        email=search.CASE_INSENSITIVE_EXACT,
                        username=search.CASE_INSENSITIVE_EXACT))

                @classmethod
                def _search_builder_mapping(self):
                    return dict(
                        id=search.ExactMatch('id'),
                        email=search.CaseInsensitiveExact('email'),
                        username=search.CaseInsensitivePrefix('username'),
                        username_exact=search.CaseInsensitiveExact('username'),
                        q=search.StringQuery())

            class TestModelBad(db.Model,
                               SearchableMixin(bucket=bucket, logger=logger)):
                pass

            if bad:
                return TestModelBad
            return TestModel
        return callback

    def setUp(self):
        self.mock_config = Mock(name='config')
        self.mock_app = Mock(name='app')
        self.mock_logger = Mock(name='logger')
        self.mock_elastic_search = Mock(name='elastic_search')
        self.mock_bucket = dict(
            elastic_search=self.mock_elastic_search,
            environment=self.tahoe.environment)
        self.mock_query = Mock(name='query')
        self.mock_filter = Mock(name='filter')
        self.mock_query.filter = self.mock_filter

        class ModelClass(object):
            query = self.mock_query

        self.mock_db = Mock(name='db')
        self.mock_db.Model = ModelClass

        ModelRegistry.destroy_all()

    def test_searchable_mixin(self):
        """
        SearchableMixin: Mixin works and adds all the right methods
        """
        ModelRegistry.register()(self.make_callback())
        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()
        assert 'TestModel' in mr.models
        assert mr.TestModel

        assert self.mock_db.Column.call_count == 0
        assert hasattr(mr.TestModel, 'query')

        assert hasattr(mr.TestModel, '_search_geo_configure')
        assert hasattr(mr.TestModel, '_search_geo_annotate_distance')
        assert hasattr(mr.TestModel, 'search')
        assert hasattr(mr.TestModel, '_search_doc_type')
        assert hasattr(mr.TestModel, '_search_index_alias')
        assert hasattr(mr.TestModel, '_search_index_name')
        assert hasattr(mr.TestModel, '_search_index')
        assert hasattr(mr.TestModel, '_search_create_index')
        assert hasattr(mr.TestModel, '_search_index_all_records')

        tm = mr.TestModel()

        assert hasattr(tm, 'index')
        assert hasattr(tm, 'can_index')
        assert hasattr(tm, 'remove_from_index')

    def test_searchable_geo_configure_without_latlng(self):
        """
        SearchableMixin: Geographical search doesn't happen unless lat/lng as specified
        """
        ModelRegistry.register()(self.make_callback())
        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()
        assert 'TestModel' in mr.models
        assert mr.TestModel

        assert self.mock_db.Column.call_count == 0
        assert hasattr(mr.TestModel, 'query')

        search_terms = dict(sorted_by='distance')
        distance = 10
        filters = Mock(name='search_filters')
        sort = Mock(name='sort')

        assert hasattr(mr.TestModel, '_search_geo_configure')
        sort, geo_search = mr.TestModel._search_geo_configure(
            search_terms,
            distance,
            filters,
            sort)

        assert not geo_search
        assert not isinstance(sort, list)
        assert not filters.append.called

    def test_searchable_geo_configure_with_latlng(self):
        """
        SearchableMixin: A query can be configured happen with a geographical centerpoint and radius
        """
        ModelRegistry.register()(self.make_callback())
        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()
        assert 'TestModel' in mr.models
        assert mr.TestModel

        assert self.mock_db.Column.call_count == 0
        assert hasattr(mr.TestModel, 'query')

        mock_lat = '-47.2929302983'
        mock_lng = '37.928392939'
        search_terms = dict(
            lat=mock_lat,
            lng=mock_lng,
            sorted_by='distance')
        distance = 10
        filters = Mock(name='search_filters')
        sort = Mock(name='sort')
        geo_filter = Mock(name="geo_filter")

        assert hasattr(mr.TestModel, '_search_geo_configure')
        sort, geo_search = mr.TestModel._search_geo_configure(
            search_terms,
            distance,
            filters,
            sort,
            geo_filter=geo_filter)

        assert geo_search
        assert filters.append.called
        geo_filter.assert_called_once_with(
            'location',
            "{},{}".format(mock_lat, mock_lng),
            distance,
            distance_type='arc',
            distance_unit='km')

        assert isinstance(sort, list)
        assert len(sort) == 1
        geo_distance_sort = sort[0].get('_geo_distance')
        assert geo_distance_sort
        assert geo_distance_sort.get('location') == "{},{}".format(mock_lat, mock_lng)
        assert geo_distance_sort.get('order') == 'asc'
        assert geo_distance_sort.get('unit') == 'km'

    def test_search_geo_annotate(self):
        """
        SearchableMixin: Results can be annotated with their geographical distance
        """
        class Meta(object):
            def __init__(self, distance):
                self.sort = [distance]

        class Result(dict):
            def __init__(self, distance):
                self._meta = Meta(distance)

        distances = [1, 1, 2, 3, 5, 8, 13]
        results = []
        for d in distances:
            results.append(Result(d))

        ModelRegistry.register()(self.make_callback())
        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()

        assert hasattr(mr.TestModel, '_search_geo_annotate_distance')
        mr.TestModel._search_geo_annotate_distance(results)

        for idx, result in enumerate(results):
            assert result.get('distance') == distances[idx]
            assert result.get('distance_unit') == 'km'

    def test_search_doc_type(self):
        """
        SearchableMixin: Model document type is the name of the model
        """
        ModelRegistry.register()(self.make_callback())
        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()
        assert hasattr(mr.TestModel, '_search_doc_type')
        assert mr.TestModel._search_doc_type() == 'TestModel'

    def test_search_index_alias(self):
        """
        SearchableMixin: Model index alias contains the model name and environment
        """
        ModelRegistry.register()(self.make_callback())

        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()
        assert hasattr(mr.TestModel, '_search_index_alias')
        assert 'TestModel'.lower() in mr.TestModel._search_index_alias()
        assert self.tahoe.environment in mr.TestModel._search_index_alias()

    def test_search_index(self):
        """
        SearchableMixin: Model can return search index settings
        """
        ModelRegistry.register()(self.make_callback())

        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()
        assert hasattr(mr.TestModel, '_search_index')
        si = mr.TestModel._search_index()
        assert si
        settings = si.get('settings')
        mappings = si.get('mappings')
        assert settings and mappings
        assert settings == search.STANDARD_SETTINGS
        tm_mapping = mappings.get(mr.TestModel._search_doc_type())
        assert tm_mapping

    def test_search_index_name(self):
        """
        SearchableMixin: Model can return search index name
        """
        ModelRegistry.register()(self.make_callback())
        suffix = "1234567890"
        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()
        assert hasattr(mr.TestModel, '_search_index_name')
        index_name = mr.TestModel._search_index_name(suffix)
        assert index_name
        assert index_name.endswith(suffix)
        assert mr.TestModel._search_index_alias() in index_name

    def test_unconfigured_builder_mapping_and_index(self):
        """
        SearchableMixin: Model throws not implemented error if the mapping or index is not defined
        """
        ModelRegistry.register()(self.make_callback(bad=True))

        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()
        assert hasattr(mr.TestModelBad, '_search_builder_mapping')
        assert hasattr(mr.TestModelBad, '_search_index_mapping')
        with self.assertRaises(NotImplementedError):
            mr.TestModelBad._search_builder_mapping()
        with self.assertRaises(NotImplementedError):
            mr.TestModelBad._search_index_mapping()

    def test_search_create_index(self):
        """
        SearchableMixin: Can create index
        """
        suffix = '1234567890'
        ModelRegistry.register()(self.make_callback())

        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()
        assert hasattr(mr.TestModel, '_search_create_index')
        index_name = mr.TestModel._search_index_name(suffix=suffix)
        index_settings = mr.TestModel._search_index().get('settings')
        mapping = mr.TestModel._search_index_mapping()
        doc_type = mr.TestModel._search_doc_type()
        mr.TestModel._search_create_index(suffix=suffix)

        assert self.mock_logger.info.call_count == 2

        self.mock_elastic_search.indices.delete_index_if_exists.assert_called_once_with(index_name)
        self.mock_elastic_search.indices.create_index.assert_called_once_with(
            index_name,
            settings=index_settings)
        self.mock_elastic_search.indices.put_mapping.assert_called_once_with(
            doc_type,
            mapping,
            [index_name])



"""
The MIT License (MIT)

Copyright (c) 2013 Lunar Technology Corporation

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from fixture import SQLAlchemyFixture


class FixturesRegistry(object):
    fixture_callbacks = []
    fixtures = {}

    def __init__(self, models, db):
        self.models = models
        self.db = db
        self._data = None
        self._dbfixture = None

    @classmethod
    def register(cls, name):
        def inner(callback):
            cls.fixture_callbacks.append((name, callback))
        return inner

    def setup(self):
        for fixture_name, fixture_callback in self.fixture_callbacks:
            fixture_class = fixture_callback(
                fixtures=self,
                models=self.models
                )
            self.fixtures[fixture_name] = fixture_class

    def load(self):
        self.data.setup()
        self.dispose()

    def dispose(self):
        self.dbfixture.dispose()

    @property
    def data(self):
        if not self._data:
            # Maintain the order these were imported in.
            fixture_classes = [self.fixtures.get(fixture_name)
                               for fixture_name, fixture_callback
                               in self.fixture_callbacks]
            self._data = self.dbfixture.data(*fixture_classes)
        return self._data

    @property
    def dbfixture(self):
        if not self._dbfixture:
            self._dbfixture = SQLAlchemyFixture(
                env=self.fixture_environment,
                engine=self.db.engine)
        return self._dbfixture

    @property
    def fixture_environment(self):
        env = dict()
        for fixture_name, fixture_class in self.fixtures.iteritems():
            env[fixture_class.__name__] = fixture_class.__model__
        return env

    def get_fixtures(self, fixture_name):
        return self.fixtures.get(fixture_name)

    def __getattr__(self, key):
            return self.get_fixtures(key)

    def __repr__(self):
        return u"<FixturesRegistry fixtures='{}' />".format(self.fixtures.keys())



__version__ = "1.0"
__author__ = "Lunar Technology Corporation"

VERSION = __version__
AUTHOR = __author__

from . import aws
from . import encoding
from . import exceptions
from . import fixtures
from . import forms
from . import misc
from . import models
from . import queries
from . import search
from . import sessions
from . import tasks
from . import tests
from . import views



import datetime

from tahoe.misc import cleanup_dict
from tahoe.misc import normalize_email_address
from tahoe.misc import url_is_ok
from tahoe.misc import get_timestamp

from tahoe.tests import TahoeTestCase


class TestMiscellaneous(TahoeTestCase):
    def test_normalize_email_address_non_gmail_with_plus(self):
        """
        Can normalize a non-Gmail address containing a plus symbol.
        """
        email_address = 'clintecker+somethingelse@example.com'
        norm_email_address = normalize_email_address(email_address)
        assert norm_email_address != email_address
        assert '+' not in norm_email_address
        assert norm_email_address == 'clintecker@example.com'

    def test_normalize_email_address_non_gmail_with_dots(self):
        """
        Can normalize a non-Gmail address containing dot symbols.
        """
        email_address = 'clint.ecker@example.com'
        norm_email_address = normalize_email_address(email_address)
        assert norm_email_address == email_address

    def test_normalize_email_address_non_gmail_with_dots_and_plus(self):
        """
        Can normalize a non-Gmail address containing both dot and plus symbols.
        """
        email_address = 'clint.ecker+something@example.com'
        norm_email_address = normalize_email_address(email_address)
        assert norm_email_address != email_address
        assert '+' not in norm_email_address
        assert norm_email_address == 'clint.ecker@example.com'

    def test_normalize_email_address_gmail_without_dots(self):
        """
        Can normalize a Gmail address not containing dot symbols.
        """
        email_address = 'clint@gmail.com'
        norm_email_address = normalize_email_address(email_address)
        assert norm_email_address == email_address

    def test_normalize_email_address_gmail_with_dots(self):
        """
        Can normalize a Gmail address containing dot symbols.
        """
        email_address = 'clint.ecker@gmail.com'
        norm_email_address = normalize_email_address(email_address)
        assert norm_email_address != email_address
        assert norm_email_address == 'clintecker@gmail.com'

    def test_normalize_email_address_gmail_with_dots_and_plus(self):
        """
        Can normalize a Gmail address containing both dot and plus symbols.
        """
        email_address = 'clint.ecker+something@gmail.com'
        norm_email_address = normalize_email_address(email_address)
        assert norm_email_address != email_address
        assert '+' not in norm_email_address
        assert norm_email_address == 'clintecker@gmail.com'

    def test_normalize_email_address_gmail_without_dots_and_plus(self):
        """
        Can normalize a Gmail address containing plus symbols.
        """
        email_address = 'clintecker+something@gmail.com'
        norm_email_address = normalize_email_address(email_address)
        assert norm_email_address != email_address
        assert '+' not in norm_email_address
        assert norm_email_address == 'clintecker@gmail.com'

    def test_url_is_ok(self):
        """
        Can validate that a URL points to a page on a given domain.
        """
        valid_domain = 'example.com'
        suspect_url = 'http://{}/something/12345?do_it=true#cool'.format(valid_domain)
        assert url_is_ok(suspect_url, valid_domain)

    def test_url_is_ok_subdomain(self):
        """
        Considers a URL containing a sub-domain of the given domain to be invalid.
        """
        valid_domain = 'example.com'
        suspect_url = 'http://www.{}/something/12345?do_it=true#cool'.format(valid_domain)
        assert not url_is_ok(suspect_url, valid_domain)

    def test_url_is_ok_no_domain_given(self):
        """
        Considers any URL as valid when no domain is given.
        """
        valid_domain = 'example.com'
        suspect_url = 'http://www.{}/something/12345?do_it=true#cool'.format(valid_domain)
        assert url_is_ok(suspect_url)

    def test_url_is_ok_relative(self):
        """
        Considers any relative URL as valid, regardless of the given domain.
        """
        valid_domain = 'example.com'
        suspect_url = '/something/12345?do_it=true#cool'
        assert url_is_ok(suspect_url, valid_domain)

    def test_url_is_ok_relative_no_domain_given(self):
        """
        Considers any relative URL as valid even if no domain is given.
        """
        suspect_url = '/something/12345?do_it=true#cool'
        assert url_is_ok(suspect_url)

    def test_cleanup_dict(self):
        """
        Cleanup dict works
        """
        d = dict(
            one=1,
            two=2,
            three=3)
        initial_keys = set(d.keys())
        keys_to_remove = ['three', 'one']
        keys_to_preserve = initial_keys.difference(set(keys_to_remove))
        cleanup_dict(d, keys_to_remove)
        for key in keys_to_remove:
            assert key not in d
        for key in keys_to_preserve:
            assert key in d

    def test_cleanup_dict_with_nonexistent_keys(self):
        """
        Cleanup dict works even when you present it with keys that don't exist
        """
        d = dict(
            one=1,
            two=2,
            three=3)
        initial_keys = set(d.keys())
        keys_to_remove = ['three', 'one', 'four']
        keys_to_preserve = initial_keys.difference(set(keys_to_remove))
        cleanup_dict(d, keys_to_remove)
        for key in keys_to_remove:
            assert key not in d
        for key in keys_to_preserve:
            assert key in d

    def test_get_timestamp(self):
        """
        Can get a timestamp of right now
        """
        before = get_timestamp(round=False)
        now = get_timestamp(round=False)
        after = get_timestamp(round=False)
        assert before < now < after

    def test_get_timestamp_of_exact_datetime(self):
        """
        Can get a timestamp of a specific datetime
        """
        dt1 = datetime.datetime(1970, 1, 1, 0, 0, 0)
        ts1 = get_timestamp(now=dt1)

        dt2 = datetime.datetime(2014, 5, 13, 16, 53, 20)
        ts2 = get_timestamp(now=dt2)

        dt3 = datetime.datetime(2038, 1, 19, 3, 14, 8)
        ts3 = get_timestamp(dt3)

        assert ts1 == 0
        assert ts2 == 1400000000
        assert ts3 == 2**31



"""
The MIT License (MIT)

Copyright (c) 2013 Lunar Technology Corporation

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
import urlparse
from functools import wraps
from werkzeug.local import LocalProxy
from flask import current_app, Response, request, _request_ctx_stack
from flask.ext.principal import Identity, identity_changed
from tahoe.encoding import json_dumps, json_handler


class ViewCollection(object):
    def __init__(self, prefix, app, version=1):
        if not prefix.endswith('/'):
            prefix += '/'
        self.prefix = prefix
        self.app = app
        self.version = version

    def add_route(self, relative_rule, **kwargs):
        def inner_function(view_function):
            use_prefix = kwargs.pop('use_prefix', True)
            use_version = kwargs.pop('use_version', True)

            if use_prefix:
                rule = urlparse.urljoin(self.prefix, relative_rule)
            else:
                rule = relative_rule

            if use_version:
                version_prefix = 'v{}/'.format(self.version)
                rule = urlparse.urljoin(version_prefix, rule)

            if not rule.startswith('/'):
                rule = '/' + rule

            self.app.add_url_rule(
                rule=rule,
                view_func=view_function,
                **kwargs)
        return inner_function


class ViewsRegistry(object):
    """
    A place to keep track of registered views and facilities to configure them
    and supplant them with application references.
    """
    view_callbacks = []

    def __init__(self, app, config, db, models, bucket, logger):
        self.db = db
        self.config = config
        self.app = app
        self.models = models
        self.bucket = bucket
        self.logger = logger

    @classmethod
    def register(cls, prefix='', version=1):
        """
        Registers a view and adds it to a list of views.  This defers the
        actualsetup of the views until we want to do it, and allows us to
        pass in important app objects that the views can use.
        """
        def inner(callback):
            options = dict(
                view_callback=callback,
                prefix=prefix,
                version=version)
            cls.view_callbacks.append(options)
        return inner

    def setup(self):
        """
        Takes the list of registered views and sets them up as the url rules
        that flask expects.

        Each view is wrapped in a closure that passes in applicaiton specific
        stuff that a view might need like a reference to the app object, the
        db, and the configuration.
        """
        for view_cb in self.view_callbacks:
            view_collection_callback = view_cb.get('view_callback')
            view_collection_prefix = view_cb.get('prefix')
            view_collection_version = view_cb.get('version')

            collection = ViewCollection(
                prefix=view_collection_prefix,
                app=self.app,
                version=view_collection_version)

            view_collection_callback(
                collection,
                db=self.db,
                app=self.app,
                config=self.config,
                models=self.models,
                bucket=self.bucket,
                logger=self.app.logger)

_security = LocalProxy(lambda: current_app.extensions['security'])

_default_unauthorized_html = """
    <h1>Unauthorized</h1>
    <p>The server could not verify that you are authorized to access the URL
    requested. You either supplied the wrong credentials (e.g. a bad
    password), or your browser doesn't understand how to supply the
    credentials required.</p>
    """


def _get_unauthorized_response(text=None, headers=None):
    text = text or _default_unauthorized_html
    headers = headers or {}
    return Response(text, 401, headers)


def _check_http_auth():
    if request.authorization:
        username = request.authorization.username
        password = request.authorization.password
        # Flask-Security provides something like this, but for some reason
        # looks up the user by their email (huh?) so I do it here by
        # username (Yuugh!), I also poke the authenticated user
        # into the request ctx so they appear as logged in via the
        # `current_user` proxy.
        user = _security.datastore.find_user(username=username)
        if user and user.verify_password(password):
            app = current_app._get_current_object()
            identity_changed.send(app, identity=Identity(user.id))
            ctx = _request_ctx_stack.top
            ctx.user = user
            return True
    return False


def http_auth_required(realm):
    """Decorator that protects endpoints using Basic HTTP authentication.
    The username should be set to the user's email address.

    :param realm: optional realm name"""

    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            if _check_http_auth():
                return fn(*args, **kwargs)
            r = _security.default_http_auth_realm if callable(realm) else realm
            h = {'WWW-Authenticate': 'Basic realm="%s"' % r}
            return _get_unauthorized_response(headers=h)
        return wrapper

    if callable(realm):
        return decorator(realm)
    return decorator


def json_response(output, status_code=200):
    if isinstance(output, basestring):
        output = dict(message=output)
    output = dict(r=output)
    response = (
        json_dumps(output, default=json_handler),
        status_code,
        {'Content-Type': 'application/json'}
        )
    return response


def paginate(query, page=None, per_page=None):
    """
    Paginates a passed in query object and returns the results.

    * By default this uses `page`=1 and the `per_page` amount set in the
      config.
    * If `page` or `per_page` are passed in via querystring, they will be
      used.
    * All of this can be overidden via the kwargs `page` and `per_page`
    """

    if not page:
        page = int(request.args.get('page', 1))

    if not per_page:
        default_per_page = 12  # TODO: pull from config
        per_page = int(request.args.get('per_page', default_per_page))

    if page < 1 or per_page < 1:
        return []

    return query.paginate(page, per_page=per_page, error_out=False).items



"""
The MIT License (MIT)

Copyright (c) 2013 Lunar Technology Corporation

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
import datetime


def TimestampMixin(db):
    class TimestampMixin(object):
        """
        Makes a model trackable by time.

        Adds ``created_at``, ``updated_at``, and ``deleted_at`` fields
        to a model and sets up all the tracking.

        Also allows for a model to be soft deleted.

        Use the ``query_active()`` method in place of the ``query`` object
        when querying for items to only get back items which have not
        been soft deleted.
        """
        NOW_FN = datetime.datetime.utcnow
        created_at = db.Column(db.DateTime, default=NOW_FN)
        updated_at = db.Column(db.DateTime, default=NOW_FN, onupdate=NOW_FN)
        deleted_at = db.Column(db.DateTime)

        @classmethod
        def query_active(cls):
            return cls.query.filter_by(deleted_at=None)

        def is_deleted(self):
            if self.deleted_at:
                return True
            return False

        def mark_deleted(self):
            self.deleted_at = self.NOW_FN()

        def mark_undeleted(self):
            self.deleted_at = None

    return TimestampMixin



from flask.ext.security.core import UserMixin
from flask.ext.security.registerable import encrypt_password
from flask.ext.security.recoverable import update_password as _update_password
from flask.ext.security.recoverable import generate_reset_password_token \
    as _generate_reset_password_token
from flask.ext.security.recoverable import reset_password_token_status \
    as _reset_password_token_status
from flask.ext.security import utils as security_utils
from flask.ext.security import signals as security_signals

from tahoe import search
from tahoe.exceptions import AccountTaken


from tahoe.models import ModelRegistry
from tahoe.models.mixins.timestampable import TimestampMixin
from tahoe.models.mixins.searchable import SearchableMixin


NOT_ENOUGH_INFO = "Please provide a username, password, and email address"
ALREADY_TAKEN = "An account is already using this username or email address."


@ModelRegistry.register()
def create_account_roles(db, models, app, config, logger, bucket):
    class AccountRole(db.Model, TimestampMixin(db)):
        """
        A mapping of Accounts to Roles. Represents the Roles associated
        with an account.
        """
        __tablename__ = 'account_roles'
        account_id = db.Column(db.Integer, primary_key=True)
        role_id = db.Column(db.Integer, primary_key=True)

        def __repr__(self):
            return "<AccountRole(account_id='{}', role_id='{}')>".format(
                self.account_id,
                self.role_id)

    return AccountRole


@ModelRegistry.register()
def create_accounts(db, models, app, config, logger, bucket):
    searchable_mixin = SearchableMixin(bucket, logger)
    timestamp_mixin = TimestampMixin(db)

    class Account(db.Model, UserMixin, searchable_mixin, timestamp_mixin):
        """
        Represents an Account.
        """
        __tablename__ = 'accounts'
        id = db.Column(db.Integer, primary_key=True)

        username = db.Column(db.String(255))
        email = db.Column(db.String(255))
        password = db.Column(db.String(255))

        active = db.Column(db.Boolean(), default=True)
        last_login_at = db.Column(db.DateTime)
        current_login_at = db.Column(db.DateTime)
        last_login_ip = db.Column(db.String(255))
        current_login_ip = db.Column(db.String(255))
        login_count = db.Column(db.Integer, default=0)

        roles = db.relationship(
            "Role",
            secondary=models.AccountRole.__table__,
            primaryjoin="and_(Account.id==AccountRole.account_id, \
                AccountRole.deleted_at==None)",
            secondaryjoin="and_(Role.id==AccountRole.role_id, \
                Role.deleted_at==None)",
            backref='accounts')

        @classmethod
        def create(cls, **kwargs):
            """
            A wrapper around the Flask-Security's concept for creating
            accounts, does all the jazz of encrypting the password
            and so forth.
            """
            username = kwargs.get('username')
            password = kwargs.get('password')
            email = kwargs.get('email')

            if not (username and password and email):
                raise ValueError(NOT_ENOUGH_INFO)
            if cls.is_valid_username_and_password(username, password):
                if cls.get_from_username_or_email(username=username,
                                                  email=email):
                    raise AccountTaken(ALREADY_TAKEN)
                kwargs['password'] = encrypt_password(kwargs['password'])
                user_datastore = bucket['user_datastore']()
                return user_datastore.create_user(**kwargs)

        @classmethod
        def get_from_username_or_email(cls, username=None, email=None):
            if username:
                account_using_username = cls.get_from_username(username)
                if account_using_username:
                    return account_using_username
            if email:
                account_using_email = cls.get_from_email(email)
                if account_using_email:
                    return account_using_email

        @classmethod
        def get_from_id(cls, id):
            """
            Finds and returns an account associated with an id
            """
            return cls.query_active().filter_by(id=id).first()

        @classmethod
        def get_from_email(cls, email):
            """
            Finds and returns an account associated with a given email address.
            """
            return cls.query_active().filter_by(email=email).first()

        @classmethod
        def get_from_username(cls, username):
            """
            Finds and returns an account associated with a given username.
            """
            return cls.query_active().filter_by(username=username).first()

        @classmethod
        def reset_password_token_status(cls, token):
            """
            Accepts a token generated for resetting a password and
            validates it.

            Returns a tuple of the form:

                (bool:expired, bool:invalid, Account:user)

            - If the token is valid but has expired, `expired` will be True
            - If the token is invalid `invalid` will be True
            - If the token is valid and has not expired, both `expired` and
              `invalid` will be False and the third item of the tuple will be
              the account associated with the token.
            """
            return _reset_password_token_status(token)

        @classmethod
        def _search_index_mapping(cls):
            return dict(properties=dict(
                id=search.EXACT_MATCH_FIELD,
                email=search.CASE_INSENSITIVE_EXACT,
                username=search.CASE_INSENSITIVE_EXACT))

        @classmethod
        def _search_builder_mapping(self):
            return dict(
                id=search.ExactMatch('id'),
                email=search.CaseInsensitiveExact('email'),
                username=search.CaseInsensitivePrefix('username'),
                username_exact=search.CaseInsensitiveExact('username'),
                q=search.StringQuery())

        def __repr__(self):
            fmt = "<Account(id='{}', username='{}', email='{}', deleted='{}')>"
            return fmt.format(
                self.id,
                self.username,
                self.email)

        def generate_reset_password_token(self):
            return _generate_reset_password_token(self)

        def send_reset_password_instructions(self):
            """
            Sends the reset password instructions email for the specified user.

            This is copied out of flask.ext.security.recoverable and modified
            for our own purposes.

            :param user: The user to send the instructions to
            """
            token = self.generate_reset_password_token()
            security_utils.send_mail(
                security_utils.config_value('EMAIL_SUBJECT_PASSWORD_RESET'),
                self.email,
                'reset_instructions',
                user=self,
                reset_token=token)
            security_signals.reset_password_instructions_sent.send(
                dict(user=self, token=token),
                app=app)

        def update_password(self, new_password):
            """
            Hashes the plaintext `new_password` into a format suitable for
            storage in a database.

            See http://pythonhosted.org/Flask-Security/configuration.html for
            configuring how passwords are hashed.
            """
            return _update_password(self, new_password)

        def verify_password(self, password):
            return security_utils.verify_password(password, self.password)

        def add_role(self, role_name):
            """
            Adds a role to this user.
            """
            role = models.Role.query_active().filter_by(name=role_name).first()
            if role:
                return bucket['user_datastore']().add_role_to_user(self, role)
            return False

        def remove_role(self, role_name):
            """
            Removes a role from this user.
            """
            role = models.Role.query_active().filter_by(name=role_name).first()
            if not role:
                return False
            return bucket['user_datastore']().remove_role_from_user(
                self,
                role_name)

        def to_dict(self):
            """
            The representation of this model when jsonified in an API response.
            """
            return dict(
                id=self.id,
                username=self.username,
                email=self.email,
                created_at=self.created_at,
                updated_at=self.updated_at,
                deleted_at=self.deleted_at)

    return Account



"""
The MIT License (MIT)

Copyright (c) 2013 Lunar Technology Corporation

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
import urlparse

from boto.s3.connection import S3Connection
from boto.s3.key import Key
from boto.exception import S3ResponseError


class S3Bucket(object):
    def __init__(self, bucket):
        """
        Accepts a boto `Bucket` instance.
        """
        self.bucket = bucket

    @property
    def name(self):
        """
        The name of this bucket.
        """
        return self.bucket.name

    def clear(self):
        # NOOP
        pass

    def get_url_for(self, key, ttl=None):
        k = Key(self.bucket)
        k.key = key
        if ttl:
            return k.generate_url(expires_in=self.url_ttl)
        return k.generate_url()

    def list(self, prefix=''):
        """
        The contents of this bucket.
        """
        keys = self.bucket.list(prefix=prefix)
        return [key.name for key in keys]

    def get_key(self, key):
        """
        Returns the value found at key.
        """
        k = Key(self.bucket)
        k.key = key
        try:
            data = k.get_contents_as_string()
        except S3ResponseError:
            data = None
        return data

    def set_key(self, key, value, replace=True):
        """
        Applies the value to the given key.
        """
        k = Key(self.bucket)
        k.key = key
        return k.set_contents_from_string(value, replace=replace)

    def copy(self, src_key, dst_bucket, dst_key, metadata=None):
        """
        - `src_key`: A string of where the data exists in the source bucket
        - `dst_bucket`: An `S3Bucket` instance where the key should be copied
                        to.
        - `dst_key`: A string of where the data should exist in the
                     destination bucket.
        - `metadata`: Metadata to be associated with new key. If metadata is
                      supplied, it will replace the metadata of the source
                      key being copied. If no metadata is supplied, the
                      source key's metadata will be copied to the new key.
        """
        s_key = Key(self.bucket)
        s_key.key = src_key
        s_key.copy(dst_bucket.name, dst_key, metadata=metadata)


class MockS3Bucket(object):
    def __init__(self, name):
        self.store = {}
        self.host = "testing.s3.amazonaws.com"
        self.name = name

    def clear(self):
        self.store = {}

    def get_url_for(self, key):
        return urlparse.urljoin("http://{}".format(self.host), key)

    def list(self, prefix=''):
        if not prefix:
            return self.store.keys()
        else:
            return [k for k in self.store.keys() if k.startswith(prefix)]

    def get_key(self, key):
        return self.store.get(key)

    def set_key(self, key, value):
        self.store[key] = value

    def copy(self, src_key, dst_bucket, dst_key, metadata=None):
        s_key_value = self.get_key(src_key)
        if s_key_value:
            dst_bucket.set_key(dst_key, s_key_value)


class AWS(object):
    def __init__(self, config):
        self.config = config
        self.aws_key_id = self.config.get('AWS_ACCESS_KEY_ID')
        self.aws_key_secret = self.config.get('AWS_ACCESS_KEY_SECRET')
        self._s3_connection = None
        self._s3_buckets = {}
        self.testing = self.config.get('TESTING', False)

    @property
    def s3_connection(self):
        if not self._s3_connection:
            self._s3_connection = self._get_s3_connection()
        return self._s3_connection

    def _get_s3_connection(self):
        return S3Connection(self.aws_key_id, self.aws_key_secret)

    def clear_mock_buckets(self):
        if self.testing:
            for bucket in self._s3_buckets.values():
                bucket.clear()

    def get_bucket(self, bucket_name):
        bucket = self._s3_buckets.get(bucket_name)
        if not bucket:
            if self.testing:
                bucket = MockS3Bucket(name=bucket_name)
            else:
                bucket = S3Bucket(
                    bucket=self.s3_connection.get_bucket(bucket_name)
                    )
            self._s3_buckets[bucket_name] = bucket
        return bucket



"""
The MIT License (MIT)

Copyright (c) 2013 Lunar Technology Corporation

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from . import mixins


class ModelRegistry(object):
    model_callbacks = []
    models = {}

    def __init__(self, app, config, db, logger, bucket):
        """
        A place to gather together all our models for easy access and mass operations
        """
        self.app = app
        self.config = config
        self.db = db
        self.logger = logger
        self.bucket = bucket

    @classmethod
    def destroy_all(cls):
        cls.model_callbacks = []
        cls.models = {}

    @classmethod
    def register(cls):
        def inner(callback):
            cls.model_callbacks.append(callback)
        return inner

    def __getattr__(self, key):
        model_class = self.models.get(key)
        if not model_class:
            raise ValueError('No such model registered as {}'.format(key))
        return model_class

    def _setup_model(self, callback):
        model_class = callback(
            config=self.config,
            db=self.db,
            models=self,
            app=self.app,
            logger=self.logger,
            bucket=self.bucket)
        return model_class.__name__, model_class

    def _update_models(self, model_name, model_class):
        self.models[model_class.__name__] = model_class

    def setup(self):
        for callback in self.model_callbacks:
            model_name, model_class = self._setup_model(callback)
            self._update_models(model_name, model_class)

    def __repr__(self):
        return u"<ModelRegistry models='{}' />".format(self.models.keys())



import subprocess
import logging

from werkzeug.contrib.cache import MemcachedCache
import ultramemcache

from flask import Flask
from flask.ext.sqlalchemy import SQLAlchemy
from flask.ext.security import Security, SQLAlchemyUserDatastore
from flask.ext.mail import Mail

import pyes
import celery

from tahoe.models import ModelRegistry
from tahoe.views import ViewsRegistry
from tahoe.tasks import TaskRegistry

from tahoe.sessions import ItsdangerousSessionInterface
from tahoe.aws import AWS

# Register all the views, models, and tasks
from models import *
from views import *
from tasks import *


class TestTahoeApplication(object):
    """
    A minimal Tahoe Application for testing
    """
    def __init__(self):
        self.config_name = 'tests.app.config.Testing'
        self.environment = 'testing'
        self._app = None
        self._db = None
        self._views = None
        self._models = None
        self._user_datastore = None
        self._security = None
        self._mail = None
        self._aws = None
        self._version = None
        self._elastic_search = None
        self._celery = None
        self._tasks = None
        self._cache = None

    @property
    def cache(self):
        if not self._cache:
            self._cache = self._get_cache()
        return self._cache

    @property
    def celery(self):
        if not self._celery:
            self._celery = self._get_celery()
        return self._celery

    @property
    def tasks(self):
        if not self._tasks:
            self._tasks = self._setup_tasks()
        return self._tasks

    @property
    def elastic_search(self):
        if not self._elastic_search:
            self._elastic_search = self._get_elastic_search()
        return self._elastic_search

    @property
    def version(self):
        if not self._version:
            self._version = self._get_version()
        return self._version

    @property
    def aws(self):
        if not self._aws:
            self._aws = self._get_aws()
        return self._aws

    @property
    def app(self):
        if not self._app:
            self._app = self._setup_app()
        return self._app

    @property
    def config(self):
        return self.app.config

    @property
    def db(self):
        if not self._db:
            self._db = self._setup_db()
        return self._db

    @property
    def models(self):
        if not self._models:
            self._models = self._setup_models()
        return self._models

    def get_user_datastore(self):
        if not self._user_datastore:
            self._user_datastore = self._setup_user_datastore()
        return self._user_datastore

    @property
    def security(self):
        if not self._security:
            self._security = self._setup_security()
        return self._security

    @property
    def mail(self):
        if not self._mail:
            self._mail = self._setup_mail()
        return self._mail

    @property
    def views(self):
        if not self._views:
            self._views = self._setup_views()
        return self._views

    @property
    def views_bucket(self):
        return dict(
            aws=self.aws,
            environment=self.environment,
            tasks=self.tasks,
            cache=self.cache)

    @property
    def model_bucket(self):
        return dict(
            user_datastore=self.get_user_datastore,
            aws=self.aws,
            elastic_search=self.elastic_search,
            environment=self.environment,
            cache=self.cache)

    @property
    def tasks_bucket(self):
        return dict(
            aws=self.aws,
            environment=self.environment,
            cache=self.cache)

    def teardown_request(self, exception=None):
        """
        Do not commit if we encounter an exception in a request.
        """
        if not exception:
            self.db.session.commit()
        if self.environment != 'Testing':
            self.db.session.close()

    def after_request(self, response):
        response.headers['Tahoe-Version'] = self.version
        return response

    def setup(self):
        """
        Execute setup routines which must be run immediately and not defferered
        """
        self.app.logger.handlers[0].setLevel(logging.WARNING)
        return [
            self.views,
            self._setup_session_interface(),
            self.security,
            self.mail,
            self.tasks]

    def _setup_session_interface(self):
        self.app.session_interface = ItsdangerousSessionInterface(
            config=self.config)

    def _setup_app(self):
        """
        Creates the Flask application and sets it up a bit
        """
        app = Flask(__name__)
        app.config.from_object(self.config_name)
        app.teardown_request(self.teardown_request)
        app.after_request(self.after_request)
        return app

    def _setup_db(self):
        """
        Initializes the database connection
        """
        return SQLAlchemy(self.app)

    def _setup_views(self):
        """
        Creates and sets up the applicaton views
        """
        views = ViewsRegistry(
            app=self.app,
            db=self.db,
            config=self.config,
            models=self.models,
            bucket=self.views_bucket,
            logger=self.app.logger)
        views.setup()
        return views

    def _setup_models(self):
        """
        Configures and sets up all of the registered models
        """
        models = ModelRegistry(
            app=self.app,
            db=self.db,
            config=self.config,
            logger=self.app.logger,
            bucket=self.model_bucket)
        models.setup()
        return models

    def _setup_user_datastore(self):
        """
        Configures the "user_datastore" used by Flask-Security
        """
        return SQLAlchemyUserDatastore(
            self.db,
            self.models.Account,
            self.models.Role)

    def _setup_security(self):
        """
        Configures the Security routines supplied by Flask-Security
        """
        return Security(
            self.app,
            self.get_user_datastore(),
            register_blueprint=False)

    def _setup_mail(self):
        """
        Configures Flask-Mail, see options in config.py
        """
        return Mail(self.app)

    def _get_version(self):
        """
        Right now this just nabs the more recent git ref
        """
        cmd = """/usr/bin/git log -n 1 --no-merges --oneline | \
                 /usr/bin/awk '{print $1}'"""
        pr = subprocess.Popen(
            [cmd],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True)
        (out, error) = pr.communicate()
        if not error:
            return out.strip()
        return "ERROR GETTING VERSION"

    def _get_aws(self):
        return AWS(self.config)

    def _get_elastic_search(self):
        return pyes.ES(self.config.get('ELASTIC_SEARCH_SERVERS'))

    def _get_celery(self):
        c = celery.Celery()
        c.config_from_object(self.config)
        return c

    def _setup_tasks(self):
        tasks = TaskRegistry(
            models=self.models,
            db=self.db,
            logger=self.app.logger,
            config=self.config,
            app=self.app,
            celery=self.celery,
            bucket=self.tasks_bucket)
        tasks.setup()
        return tasks

    def _get_cache(self):
        server_list = self.config.get('MEMCACHED_SERVERS')
        ultramemcache_client = ultramemcache.Client(server_list)
        return MemcachedCache(ultramemcache_client)

    def drop_all_tables(self):
        self.app.logger.debug("Dropping all tables")
        self.db.drop_all()

    def create_all_tables(self):
        self.app.logger.debug("Creating all tables")
        self.db.create_all()



import os


class Testing(object):
    DEBUG = True
    TESTING = True

    SECURITY_PASSWORD_HASH = 'plaintext'
    SQLALCHEMY_DATABASE_URI = 'sqlite+pysqlite:///testing.db'
    # Forces Celery to execute tasks immediately while blocking
    CELERY_ALWAYS_EAGER = True

    SQLALCHEMY_COMMIT_ON_TEARDOWN = True
    SESSION_PROTECTION = False

    SECURITY_RESET_SALT = 'TESTING_SALT'
    SECURITY_PASSWORD_SALT = 'TESTING_SALT'
    SECRET_KEY = 'TESTING_KEY'
    SESSION_SALT = 'TESTING_SALT'

    MAIL_SERVER = 'mail.tahoe.local'
    MAIL_PORT = 587
    MAIL_DEBUG = True

    MAIL_USERNAME = 'test'
    MAIL_PASSWORD = 'test'

    MAIL_USE_TLS = True
    MAIL_USE_SSL = False
    MAIL_SUPPRESS_SEND = True
    MAIL_FAIL_SILENTLY = False
    DEFAULT_MAIL_SENDER = 'Tahoe <noreply@tahoe.local>'

    SECURITY_BLUEPRINT_NAME = ''

    AWS_ACCESS_KEY_ID = os.environ.get('AWS_ACCESS_KEY_ID')
    AWS_ACCESS_KEY_SECRET = os.environ.get('AWS_ACCESS_KEY_SECRET')

    ELASTIC_SEARCH_SERVERS = ['127.0.0.1:9200']
    MEMCACHED_SERVERS = ['127.0.0.1:11211']

    # For Task Workers
    BROKER_URL = 'redis://localhost:6379/0'
    CELERY_BROKER_TRANSPORT_OPTIONS = {
        'visibility_timeout': 3600
    }
    CELERY_RESULT_SERIALIZER = 'json'
    CELERY_ENABLE_UTC = True



from tahoe.models import ModelRegistry
from tahoe.models.mixins.timestampable import TimestampMixin
from flask_security.core import RoleMixin


@ModelRegistry.register()
def create_roles(db, models, app, config, logger, bucket):
    timestamp_mixin = TimestampMixin(db)

    class Role(db.Model, RoleMixin, timestamp_mixin):
        """
        Represents a role
        """
        __tablename__ = 'roles'
        id = db.Column(db.Integer, primary_key=True)
        name = db.Column(db.String(255), index=True)
        description = db.Column(db.String(255, convert_unicode=True))

        def __repr__(self):
            return "<Role(id='{}', name='{}')>".format(self.id, self.name)

        @classmethod
        def get_from_name(cls, name):
            return cls.query_active().filter_by(name=name).first()

    return Role



"""
Tahoe Framework
---------

Provides the building blocks for a RESTful Web Service
"""
from setuptools import setup

setup(
    name='tahoe',
    version="1.0",
    url='https://github.com/lunarcorp/tahoe',
    license='The MIT License: http://www.opensource.org/licenses/mit-license.php',
    author='Lunar Technology Corporation',
    author_email='info@ltc.io',
    description='Provides the building blocks for a RESTful Web Service',
    long_description=__doc__,
    packages=['tahoe', 'tahoe.models', 'tahoe.models.mixins'],
    zip_safe=False,
    platforms='any',
    scripts=[],
    package_data={'tahoe': ['VERSION']},
    install_requires=[
        'setuptools',
        'Flask==0.9',
        'SQLAlchemy==0.8.1',
        'Flask-SQLAlchemy',
        'Flask-Security==1.6.3',
        'py-bcrypt==0.3',
        'simplejson==3.3.0',
        'boto==2.9.5',
        'pyes==0.20.0',
        'sphinx',
        'sphinxcontrib-httpdomain',
        'python-ultramemcached==0.0.2',
        'fixture==1.4',
        'alembic==0.5.0',
        'celery==3.0.19',
        'celery-with-redis==3.0'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



"""
The MIT License (MIT)

Copyright (c) 2013 Lunar Technology Corporation

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


class EncoderMixin(object):
    """
    Provides a method for encoding a model-defined encoder and decoder chain.
    Allows you to take some kind of data, say a Python object, and serialize
    it to some form, say for storing in a database.

    Imagine if you want to encode an object into a format that can be stored
    in a database text field and be reconstitutable into a python object
    again in the future.

    You might define an encoder-decoder chain like so:

    .. code-block:: python

        encode_decode_chain = [
            (pickle.dumps, pickle.loads),
            (base64.b64encode, base64.b64decode)
            ]

    This implies that when encoding data passed into the `cls.encode`
    method, the data would first be run through `pickle.dumps` and then
    ``base64.b64encode``.

    When decoding it would work the other way around, running the encoded
    data first through ``base64.b64decode`` and then ``pickle.loads``.

    Pickle could be replaced with JSON, and the text encoding could be replaced
    with gzip encoding, for example.
    """
    encode_decode_chain = []

    @classmethod
    def encode(cls, data):
        for link in cls.encode_decode_chain:
            data = link[0](data)
        return data

    @classmethod
    def decode(cls, data):
        for link in reversed(cls.encode_decode_chain):
            data = link[1](data)
        return data



"""
The MIT License (MIT)

Copyright (c) 2013 Lunar Technology Corporation

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import pyes
from tahoe import search
from tahoe.misc import get_timestamp


def SearchableMixin(bucket, logger):
    elastic_search = bucket.get('elastic_search')
    environment = bucket.get('environment')

    class SearchableMixin(object):
        """
        Makes a Model Searchable.

        Consult the ``Account`` model to see examples.

        Essentially, all you have to do to make a model searchable
        is to include this mixin in its definition and then define two
        methods to define how you would like the model indexed by
        ElasticSearch.
        """
        @classmethod
        def _search_geo_configure(cls, search_terms, distance, filters, sort, geo_filter=pyes.GeoDistanceFilter):
            """
            Determines is the query was geo-centered and configure the query appropriately.

            Results will be geo-filtered if the query included both ``lat``
            and ``lng`` facet.  The results will be filtered in KILOMETERS
            by the ``distance`` argument.

            If there existed a ``sorted_by`` parameter set to 'distance' then
            the results will be organized by distance from the given
            lat/lng coordinates.

            If ``lat`` and ``lng`` were not specified, this is a no-op.
            """
            lat = search_terms.get('lat')
            lng = search_terms.get('lng')
            sorted_by = search_terms.get('sorted_by')
            geo_search = False
            if lat and lng:
                geo_search = True
                location = "{},{}".format(lat, lng)
                filters.append(
                    geo_filter(
                        "location",
                        location,
                        distance,
                        distance_type="arc",
                        distance_unit="km"
                        ))
                if sorted_by == 'distance':
                    sort = [dict(_geo_distance=dict(
                        location=location,
                        order="asc",
                        unit="km",
                        ))]
            return sort, geo_search

        @classmethod
        def _search_geo_annotate_distance(cls, geo_results):
            """
            Annotates the results with each result's distance from the
            latitude and longitude given in the query.
            """
            for result in geo_results:
                if result._meta.sort:
                    result[u'distance'] = result._meta.sort[0]
                    result[u'distance_unit'] = 'km'
            return geo_results

        @classmethod
        def search(cls, per_page=12, page=1, distance=100, sort=None, **search_terms):
            """
            Parameters:

            - `per_page`: How many results to return per page.
            - `page`: Which page of results to return.
            - `sort` (list): `[{'created_at':'asc'}, {'username':'desc'}]`
            - `**search_terms`: Keyword args: `{'username':'clint'}`
            """
            queries = []
            filters = []
            # Obtain the mapping as defined by the model
            mapping = cls._search_builder_mapping()
            # Build up the query using the given terms and the allowed
            # fields in the mapping.
            logger.info("Got search terms: {}".format(search_terms))
            for field, value in search_terms.iteritems():
                query_builder = mapping.get(field)
                if query_builder:
                    query = query_builder.for_value(value)
                    if query_builder.filter:
                        filters.append(query)
                    else:
                        queries.append(query)
            logger.info("Produced queries: {}".format(queries))
            # Setup geo filters if necessary
            sort, geo_search = cls._search_geo_configure(
                search_terms,
                distance,
                filters,
                sort
                )
            f = None
            if queries:
                # All queries must be satisfied
                q = pyes.BoolQuery(must=queries)
            else:
                # If no queries happened to be produced by this query,
                # we start with Matching everything.
                q = pyes.MatchAllQuery()
            if filters:
                # We AND together all the filters.
                f = pyes.ANDFilter(filters)

            try:
                page = int(page)
            except ValueError:
                page = 1

            try:
                per_page = int(per_page)
            except ValueError:
                per_page = 12

            start = (page - 1) * per_page
            # Produce a Search Object. Hint: You can print this to see
            # the JSON search object that will be transmitted to the ES
            # server.
            s = pyes.Search(
                query=q,
                filter=f,
                sort=sort,
                start=start,
                size=per_page
                )

            # Produce the resultset.
            results = elastic_search.search(
                s,
                doc_types=[cls._search_doc_type()],
                indices=[cls._search_index_alias()],
                start=start
                )

            logger.info("Got {} results: {}.".format(results.total, results))

            # Results is a lazy object, so lets chop it down.
            results = [p for p in results]
            # If the query was geo-search targeted, we will annotate each
            # result with the distance from the source coordinates (in KM).
            if geo_search:
                results = cls._search_geo_annotate_distance(results)
            return results

        @classmethod
        def _search_doc_type(cls):
            """
            Determines which "Document Type" mapping to use in ElasticSearch.

            We just use the string of the Class we're indexing.
            """
            return cls.__name__

        @classmethod
        def _search_index_alias(cls):
            """
            We use ElasticSearch aliases to point to timestamped versions
            of the Model's real index.  Makes things easier.

            Accounts would be indexed at:

                ``accounts-production``
            """
            return "{}-{}".format(
                cls._search_doc_type().lower(),
                environment.lower()
                )

        @classmethod
        def _search_index_name(cls, suffix):
            """
            The real indexes are generally something like:

                ``photos-staging-<< timestamp() >>``
            """
            return "{}-{}".format(
                cls._search_index_alias(),
                str(suffix)
                )

        @classmethod
        def _search_index(cls):
            """
            Defines how the ElasticSearch Index for a given model should be configured.
            """
            return dict(
                settings=search.STANDARD_SETTINGS,
                mappings={cls._search_doc_type(): cls._search_index_mapping()}
                )

        @classmethod
        def _search_builder_mapping(self):
            """
            How the fields of a model should be **queried**.

            Define this method on your model. It should return a mapping
            of terms you expect to be passed to your model's ``search()``
            method.

            Consult the ``Account`` model to get common examples.
            Additional builder types can be defined in ``tahoe.search``.
            """
            raise NotImplementedError("Define a search builder mapping on your Model.")

        @classmethod
        def _search_index_mapping(cls):
            """
            How to fields of a model should be **indexed**.

            Define this method on your model. It should return a mapping of
            fields present in your model's ``to_dict()`` method and how
            ElasticSearch should index those fields.

            Consult the ``Account`` model to get common examples. Additional
            index types can be defined in ``tahoe.search``.
            """
            raise NotImplementedError("Define a search index mapping on your Model.")

        @classmethod
        def _search_create_index(cls, suffix):
            """
            Creates a new timestamped index for a model.

            Produces a timestamped index for a Model and applies the
            developer-defined mapping for each field.
            """
            index_name = cls._search_index_name(suffix=suffix)
            logger.info("Creating index: {}".format(index_name))
            index_settings = cls._search_index().get('settings')
            elastic_search.indices.delete_index_if_exists(index_name)
            elastic_search.indices.create_index(
                index_name,
                settings=index_settings
                )
            mapping = cls._search_index_mapping()
            doc_type = cls._search_doc_type()
            logger.info(" - Applying mapping for: {}".format(doc_type))
            elastic_search.indices.put_mapping(doc_type, mapping, [index_name])
            return index_name

        @classmethod
        def _search_index_all_records(cls, suffix=None, index_name=None, remove_old=False, refresh_now=False):
            """
            Sets up a new index for a model and re-indexes all the data for that model.

            Goal: Online generation of a new index without index downtime.

            Procedure:

                - Create a new timestamped index for the Model
                - Index every un-deleted document into this new index.
                - Alias the commonly-known index name to the new timestamped
                  index.
                - Remove all the old indices.
            """
            suffix = suffix or get_timestamp()
            index_name = index_name or cls._search_create_index(suffix=suffix)
            index_alias = cls._search_index_alias()
            doc_type = cls._search_doc_type()
            if hasattr(cls, 'query_active'):
                query = cls.query_active()
            else:
                query = cls.query

            num_records_indexed = 0
            logger.info(' - Indexing {} records into {}.'.format(
                doc_type,
                index_name
                ))
            for record in query.all():
                if record.can_index():
                    num_records_indexed += 1
                    record.index(index_name=index_name, bulk=True)

            logger.info(' - Indexed {} {}(s) into {}'.format(
                num_records_indexed,
                doc_type,
                index_name
                ))

            if refresh_now:
                elastic_search.indices.refresh(index_name)
                logger.info(' - Refreshed index: {}'.format(index_name))

            elastic_search.indices.delete_index_if_exists(index_alias)
            elastic_search.indices.add_alias(index_alias, index_name)
            logger.info(' - Aliased {} to {}'.format(index_name, index_alias))

            if remove_old:
                logger.info(' - Removing old indices:')
                current_indices = elastic_search.indices.aliases()
                # Remove all old indexes which match this prefix (except the one we just made!)
                cleanup_indices = [a_index_name for a_index_name, _ in current_indices.iteritems() if a_index_name.startswith(index_alias) and a_index_name != index_name]
                for index_name in cleanup_indices:
                    elastic_search.indices.delete_index_if_exists(index_name)
                    logger.info("    - Deleted old index: {}".format(index_name))

            return dict(
                index_name=index_name,
                index_alias=index_alias,
                doc_type=doc_type,
                num_records_indexed=num_records_indexed,
                num_indices_deleted=len(cleanup_indices),
                indices_deleted=cleanup_indices
                )

        def index(self, index_name=None, bulk=False):
            """
            Indexes this model instance.

            Does not immediately make it available for search.  If you need
            the record to appear immediately, you must refresh the index.

            Otherwise, the results will only become available as ES
            periodically refreshes its indexes.
            """
            index_name = index_name or self.__class__._search_index_alias()
            return elastic_search.index(
                doc=self.to_dict(),
                index=index_name,
                doc_type=self.__class__._search_doc_type(),
                id=self.id,
                bulk=True
                )

        def can_index(self):
            """
            Determines if a model can be indexed.
            """
            return hasattr(self, 'index') and hasattr(self, 'to_dict')

        def remove_from_index(self, index_name=None):
            """
            Removes an instance from the search index.
            """
            index_name = index_name or self.__class__._search_index_alias()
            try:
                elastic_search.delete(
                    index=index_name,
                    doc_type=self.__class__._search_doc_type(),
                    id=self.id
                    )
            except pyes.exceptions.NotFoundException:
                return False
            return True

    return SearchableMixin



"""
The MIT License (MIT)

Copyright (c) 2013 Lunar Technology Corporation

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
import base64
from contextlib import contextmanager

from flask.ext.testing import TestCase, ContextVariableDoesNotExist
from nose.plugins.skip import SkipTest
from tahoe.fixtures import FixturesRegistry


def skip(test):
    """ @skip decorator to skip WIP tests """
    def wrapper(*args, **kwargs):
        raise SkipTest()
    wrapper.__name__ = test.__name__
    wrapper.__doc__ = test.__doc__
    return wrapper


def index_db(test):
    def wrapper(self, *args, **kwargs):
        self.reindex_db()
        test(self, *args, **kwargs)
    wrapper.__name__ = test.__name__
    wrapper.__doc__ = test.__doc__
    return wrapper


class TahoeTestCase(TestCase):
    """
    Test harness for the Tahoe application
    """
    TAHOE_APPLICATION = None

    @classmethod
    def register_tahoe_application(cls, tahoe_app):
        cls.TAHOE_APPLICATION = tahoe_app

    def __init__(self, *args, **kwargs):
        self.db = self.tahoe.db
        self.models = self.tahoe.models
        self._fixtures = None
        super(TahoeTestCase, self,).__init__(*args, **kwargs)

    @property
    def tahoe(self):
        return self.TAHOE_APPLICATION

    def create_app(self):
        return self.tahoe.app

    def _pre_setup(self):
        super(TahoeTestCase, self)._pre_setup()
        self.reset_db()

    def _post_teardown(self):
        super(TahoeTestCase, self)._post_teardown()
        self.db.session.close()

    @property
    def fixtures(self):
        if not self._fixtures:
            self._fixtures = FixturesRegistry(
                models=self.models,
                db=self.db)
            self._fixtures.setup()
        return self._fixtures

    @property
    def current_session(self):
        try:
            sess = self.get_context_variable('session')
        except ContextVariableDoesNotExist:
            sess = None
        return sess

    @property
    def logger(self):
        return self.app.logger

    def ppr(self, response, chunk=256, fuzziness=0.1):
        print "Status code: {}".format(response.status_code)
        print "Headers    : {}".format(response.headers)
        print "==============="
        print "Data:"
        if len(response.data) > (chunk * (1.0 + fuzziness)):
            print "{}\n...\n{}\n\n".format(
                response.data[0:(chunk / 2)],
                response.data[-(chunk / 2):])
        else:
            print "{}\n\n".format(response.data)

    def reset_db(self):
        self.db.engine.dispose()
        self.tahoe.drop_all_tables()
        self.tahoe.create_all_tables()
        self.fixtures.load()
        self.tahoe.aws.clear_mock_buckets()

    def reindex_db(self):
        """
        Will create new indexes and index all the data currently in the
        database for the following models. Should only be called when you
        need to start with a blank elasticsearch slate for a test.
        """
        # This shouldn't be hardcoded, right?
        models_to_index = [self.models.Account]
        indices = []
        for model in models_to_index:
            res = model._search_index_all_records(remove_old=True, refresh_now=False)
            indices.append(res['index_name'])
        self.tahoe.elastic_search.refresh(indices)

    def _parse_cookie(self, cookie_data):
        cookie = dict()
        cookie_bits = [pair.split('=') for pair in cookie_data.split('; ')]
        cookie['name'], cookie['value'] = cookie_bits.pop(0)
        for pair in cookie_bits:
            if len(pair) == 0:
                continue
            key = pair[0]
            if len(pair) == 1:
                value = True
            else:
                value = pair[1]
            cookie[key] = value
        cookie_expires = cookie.get('expires')
        if cookie_expires:
            cookie['session'] = False
        else:
            cookie['session'] = True

        if cookie_expires == 'Thu, 01-Jan-1970 00:00:00 GMT' and \
            cookie.get('Max-Age') == '0' and \
                cookie.get('value', '').strip() == '':
                    cookie['destroyed'] = True
        else:
            cookie['destroyed'] = False
        return cookie

    def _parse_cookies(self, response):
        cookies = dict()
        for key, value in response.headers:
            if key == 'Set-Cookie':
                cookie = self._parse_cookie(value)
                cookies[cookie['name']] = cookie
        return cookies

    def _get_cookie(self, response, cookie_name):
        cookies = self._parse_cookies(response)
        return cookies.get(cookie_name)

    def cookieWasSet(self, response, cookie_name, **kwargs):
        cookie = self._get_cookie(response, cookie_name)
        if cookie:
            if kwargs:
                for key, value in kwargs.iteritems():
                    if cookie.get(key) != value:
                        return False
            return True
        return False

    def cookieWasDestroyed(self, response, cookie_name):
        cookie = self._get_cookie(response, cookie_name)
        return cookie and cookie.get('destroyed')

    def basic_auth_headers(self, username, password):
        return {
            'Authorization': 'Basic ' + base64.b64encode(username +
            ":" + password)
        }

    def basic_auth_request(self, *args, **kwargs):
        valid_methods = ['get', 'post', 'put', 'delete', 'patch']

        method = kwargs.pop('method', 'get')
        if method not in valid_methods:
            raise ValueError('Please use one of the following \
                methods: {}'.format(valid_methods))
        request_method = getattr(self.client, method)
        if not request_method:
            raise ValueError('No request method for: {}'.format(method))

        account = kwargs.pop('account', None)
        if account:
            username = account.username
            password = account.password
            headers = kwargs.get('headers', {})
            self.logger.debug("Old Headers: {}".format(headers))
            auth_headers = self.basic_auth_headers(
                username=username,
                password=password
                )
            self.logger.debug("Generated auth headers: {}".format(auth_headers))
            headers.update(auth_headers)
            kwargs['headers'] = headers
            self.logger.debug("Making an authenticated request as {} with: {} || {}".format(username, args, kwargs))
        else:
            self.logger.debug("Making an unauthenticated request with: {} || {}".format(args, kwargs))
        return request_method(*args, **kwargs)

    def basic_auth_get(self, *args, **kwargs):
        kwargs['method'] = 'get'
        return self.basic_auth_request(*args, **kwargs)

    def basic_auth_post(self, *args, **kwargs):
        kwargs['method'] = 'post'
        return self.basic_auth_request(*args, **kwargs)

    def basic_auth_put(self, *args, **kwargs):
        kwargs['method'] = 'put'
        return self.basic_auth_request(*args, **kwargs)

    def basic_auth_delete(self, *args, **kwargs):
        kwargs['method'] = 'delete'
        return self.basic_auth_request(*args, **kwargs)

    def basic_auth_patch(self, *args, **kwargs):
        kwargs['method'] = 'patch'
        return self.basic_auth_request(*args, **kwargs)

    def assert201(self, response):
        self.assertStatusCode(response, status_code=201)

    def assert209(self, response):
        self.assertStatusCode(response, status_code=209)

    def assertStatusCode(self, response, status_code):
        assert self.isStatusCode(response, status_code)

    def isStatusCode(self, response, status_code):
        return response.status_code == status_code

    def get_json(self, response):
        json = response.json
        return json.get('r')

    @contextmanager
    def config_change(self, key, temp_value):
        old_value = self.app.config.get(key)
        self.logger.debug("Changing config key ({}) from '{}' to \
            '{}'.".format(key, old_value, temp_value))
        self.app.config[key] = temp_value
        yield
        if old_value:
            self.logger.debug("Resetting config key ({}) from '{}' to \
                '{}'.".format(key, temp_value, old_value))
            self.app.config[key] = old_value
        else:
            self.logger.debug("Removing config key ({}).".format(key))
            del self.app.config[key]



from tahoe.models import ModelRegistry
from tahoe.tests import TahoeTestCase
from mock import Mock


class TestModel(object):
    @classmethod
    def set_args(cls, args):
        cls.args = args

    @classmethod
    def set_kwargs(cls, kwargs):
        cls.kwargs = kwargs

    @classmethod
    def destroy(cls):
        cls.args = None
        cls.kwargs = None


class TestFoo(TestModel):
    pass


class TestBar(TestModel):
    pass

test_models = [TestFoo, TestBar]


def make_callback(test_model):
    def callback(*callback_args, **callback_kwargs):
        test_model.set_args(callback_args)
        test_model.set_kwargs(callback_kwargs)
        return test_model
    return callback


class TestModels(TahoeTestCase):
    def setUp(self):
        self.mock_config = Mock(name='config')
        self.mock_db = Mock(name='db')
        self.mock_app = Mock(name='app')
        self.mock_logger = Mock(name='logger')
        self.mock_bucket = Mock(name='bucket')

    def tearDown(self):
        ModelRegistry.destroy_all()
        for test_model in test_models:
            test_model.destroy()

    def test_registering_a_model(self):
        """
        Can register a callback
        """
        before_callbacks = len(ModelRegistry.model_callbacks)
        before_models = len(ModelRegistry.models)

        test_model = test_models[0]
        callback = make_callback(test_model)
        ModelRegistry.register()(callback)

        after_callbacks = len(ModelRegistry.model_callbacks)
        after_models = len(ModelRegistry.models)

        assert after_callbacks == before_callbacks + 1
        assert callback in ModelRegistry.model_callbacks
        assert after_models == before_models

    def test_setup_model(self):
        """
        Model can be setup as needed
        """
        before_models = len(ModelRegistry.models)

        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)

        test_model = test_models[1]
        callback = make_callback(test_model)
        model_name, model_class = mr._setup_model(callback)

        after_models = len(ModelRegistry.models)

        assert after_models == before_models
        assert model_class == test_model
        assert model_name == test_model.__name__

        assert model_class.kwargs['models'] == mr
        assert model_class.kwargs['app'] == self.mock_app
        assert model_class.kwargs['config'] == self.mock_config
        assert model_class.kwargs['db'] == self.mock_db
        assert model_class.kwargs['logger'] == self.mock_logger
        assert model_class.kwargs['bucket'] == self.mock_bucket

    def test_setup_and_add_model(self):
        """
        Model Class has app-level instances in scope when callback is setup
        """
        ModelRegistry.destroy_all()

        assert len(ModelRegistry.models) == 0
        test_model = TestFoo
        callback = make_callback(test_model)

        ModelRegistry.register()(callback)

        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()

        assert len(ModelRegistry.models) == 1
        assert 'TestFoo' in ModelRegistry.models
        assert 'TestFoo' in mr.models
        assert mr.TestFoo

        TF = mr.TestFoo

        assert TF.kwargs['models'] == mr
        assert TF.kwargs['app'] == self.mock_app
        assert TF.kwargs['config'] == self.mock_config
        assert TF.kwargs['db'] == self.mock_db
        assert TF.kwargs['logger'] == self.mock_logger
        assert TF.kwargs['bucket'] == self.mock_bucket

    def test_nonexistant_model_raises_exception(self):
        """
        Attempting to access a model that isn't register raises ValueError
        """
        ModelRegistry.destroy_all()

        assert len(ModelRegistry.models) == 0
        test_model = TestFoo
        callback = make_callback(test_model)

        ModelRegistry.register()(callback)

        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)
        mr.setup()

        assert mr.TestFoo

        with self.assertRaises(ValueError):
            mr.TestBar

    def test_registry_repr_includes_all_setup_models(self):
        """
        Setting up a model causes it to show in the repr for the registry.
        """
        ModelRegistry.destroy_all()

        assert len(ModelRegistry.models) == 0
        test_model = TestFoo
        callback = make_callback(test_model)

        ModelRegistry.register()(callback)

        mr = ModelRegistry(
            app=self.mock_app,
            config=self.mock_config,
            db=self.mock_db,
            logger=self.mock_logger,
            bucket=self.mock_bucket)

        assert 'TestFoo' not in str(mr)
        assert 'TestBar' not in str(mr)

        mr.setup()

        assert 'TestFoo' in str(mr)
        assert 'TestBar' not in str(mr)



"""
The MIT License (MIT)

Copyright (c) 2013 Lunar Technology Corporation

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
import base64
import datetime
from dateutil import parser as date_parser
import urlparse
import uuid


def cleanup_dict(d, keys=None):
    """
    Removes keys from the feed items that are not required to clients.
    """
    for k in keys:
        if k in d:
            del(d[k])


def get_timestamp(now=None, round=True):
    now = now or datetime.datetime.utcnow()
    timestamp = (now - datetime.datetime(1970, 1, 1)).total_seconds()
    if round:
        return int(timestamp)
    return timestamp


def timestamp_to_datetime(timestamp):
    return datetime.datetime.fromtimestamp(int(timestamp))


def date_string_to_datetime(date_string):
    return date_parser.parse(date_string)


def get_guid():
    return base64.urlsafe_b64encode(uuid.uuid4().bytes).rstrip('=')


def normalize_email_address(raw_email_address):
    local_part, sep, domain_part = raw_email_address.partition('@')
    local_part_real, sep, local_part_plus = local_part.partition('+')
    # Google allows any number of dots in the name while all permutations point
    # to the same account.
    if domain_part == 'gmail.com':
        local_part_real = local_part_real.replace('.', '')
    return "{}@{}".format(local_part_real, domain_part)


def url_is_ok(suspect_url, valid_domain=None):
    if valid_domain:
        return (urlparse.urlparse(suspect_url).netloc in ('', valid_domain))
    return True
